/** _besi0l - zeroth-order modified Bessel function of the first kind.
 Rev. 1.4.3 (Mar. 5, 2023) (c) 2023, Takayuki HOSODA
 http://www.finetune.co.jp/~lyuka/technote/bessel/besi0.html
 _besi0l(x) returns the 0th-order modified Bessel function of the first kind
 of real argument x.
 SX_CTRL is the threshold value for switching formulas, and the standard
 setting is 24.0. For x <= SX_CTRL, use the infinite series, which is the
 definition of the zeroth-order modified Bessel function of the first kind.
 Fox x > SX_CTRL, use a 27th order polynomial approximation.
 The maximum error for 0 <= x < 768 is 6.5e-19.
 SS_BORDER is the threshold for computing an infinite series into two partial
 sums; SS_BORDER = 512 is appropriate when SX_CTRL = 24.
*/
#include "besi0l.h"
long double _besi0l(long double x) {
    int         n, m;
    long double sum, term, d, sub;
    if (0.0L == (x = fabsl(x))) return 1.0L;
    if (x <= SX_CTRL) {
        for (n = 1, sum = term = 1.0L; term > sum / SS_BORDER; sum += term, n++)
            term *= (0.25L * x * x) / (long double)(n * n); /* Major part calculation */
        for (sub = 0.0L; term >= (LDBL_EPSILON) * sub; sub += term, n++)
            term *= (0.25L * x * x) / (long double)(n * n); /* Minor part calculation */
        return sum + sub;
    } else {
        for (n = 0, sum = d = 0.0L, term = 1.0L; n < SX_NCOEF; n++, term *= (1.0L / x))
            sum += SXI0INF[n] * term;
        m = floorl(x); d = 1.0L;
        for (n = 0; n < EXP_NMAX; n++)
            if ((1 << n) & m)
                d *= EXP_POW2N[n]; /* Range reduction for expl() */
        return d * expl(x - (long double)m) / sqrtl(x) * sum;
    }
}
#ifdef DEBUG_BESI0L
/* A compile option of -ffast-math may be needed for gcc as follows.
 * gcc -Wall -O -ffast-math -DDEBUG_BESI0L besi0l.c -o besi0 
 */
#include <stdio.h>
#include <stdlib.h>
#ifdef __FreeBSD__
#include <floatingpoint.h>
#endif
int
main(int argc, char *argv[]) {
    long double       x;
    char             *dummy;
    #ifdef __FreeBSD__
    fpsetprec(FP_PE);
    fpsetround(FP_RN);
    #endif
    if (argc == 2) {
        x = strtold(argv[1], &dummy);
        printf("besi0l(%.18Lg) ~= %18.18Lg\n", x, _besi0l(x));
    } else {
        printf("%s x\n", argv[0]);
    }
    return(0);
}
#endif
